// IndustryPackCarrier.h

/********************COPYRIGHT NOTIFICATION**********************************
  This software was developed under a United States Government license
  described on the COPYRIGHT_UniversityOfChicago file included as part
  of this distribution.
  ***************************************************************************
  */
/*
  Current Author: Thomas Coleman
  Original Author: Jim Kowalkowski (for Hideos)
  Date: 09JUL98
  */

#ifndef IndustryPackCarrierH
#define IndustryPackCarrierH

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>

#include <vxLib.h>
#include <intLib.h>

#include "DLList.h"

struct ip_id_prom
{
  unsigned short ascii1;
  unsigned short ascii2;
  unsigned short ascii3;
  unsigned short ascii4;
  unsigned short manufacturer;
  unsigned short model;
  unsigned short revision;
  unsigned short reserved;
  unsigned short driverLow;
  unsigned short driverHigh;
  unsigned short numBytesUsed;
  
  unsigned short userStuff[32-11];
};
typedef struct ip_id_prom IP_ID_PROM;

class IndustryPackSite;

class IndustryPackCarrier
{
 public:
  IndustryPackCarrier();
  int registerName(const char *carrierName);
  static IndustryPackCarrier *find(const char *carrierName);
  IndustryPackSite *findSite(const char *siteName);
  IndustryPackSite *addSite(IndustryPackSite *pSite);
  const char *getCarrierName();
  virtual volatile void *allocMemMapIP(int size) = 0;
 private:
  const char *carrierName;
  DLList *pIpSiteList;
};

class IndustryPackSite
{
 public:
  IndustryPackSite();
  const char *getSiteName();
  virtual volatile void* allocMemMapIP(int size) = 0;
  virtual volatile IP_ID_PROM* getMemBaseID() = 0;
  virtual volatile void* getMemBaseIO() = 0;
  virtual volatile void* getMemBaseIP() = 0;
  virtual int intConfig(int num) = 0;
  virtual void intEnable(int num) = 0;
  virtual void intDisable(int num) = 0;
  virtual void intClear(int num) = 0;
 protected:
  const char *siteName;
  IndustryPackCarrier *pIpCarrier;
 private:
};

#endif /* IndustryPackCarrierH */

