// Message.cc

/* Author:  Marty Kraimer Date:    26MAR1998 */
 
/********************COPYRIGHT NOTIFICATION**********************************
This software was developed under a United States Government license
described on the COPYRIGHT_UniversityOfChicago file included as part
of this distribution.
****************************************************************************/

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>

#include <semLib.h>

#define mpfTypeGLOBAL
#include "mpfType.h"
#undef mpfTypeGLOBAL

#include "Message.h"
#include "MessagePvt.h"
#include "ConnectMessage.h"
#include "FreeList.h"
#include "DLList.h"
#include "WatchDog.h"

Router *prouter = 0;
static int defaultServerQueueSize = 100;


int32 Message::toBuffer(char **buffer) const
{
    int32 itype = (int32)type;
    int32 len = 0;
    len += MPF::toBuffer(buffer,itype);
    len += MPF::toBuffer(buffer,clientType);
    len += MPF::toBuffer(buffer,clientExtra);
    return(len);
}

int32 Message::fromBuffer(const char **buffer)
{
    int32 nbytes=0;
    int32 itype;
    nbytes += MPF::fromBuffer(buffer,&itype);
    type = (messageType)itype;
    nbytes += MPF::fromBuffer(buffer,&clientType);
    nbytes += MPF::fromBuffer(buffer,&clientExtra);
    return(nbytes);
}

int32 Message::fromBufferSwitch(const char **buffer)
{
    int32 nbytes=0;
    int32 itype;
    nbytes += MPF::fromBufferSwitch(buffer,&itype);
    type = (messageType)itype;
    nbytes += MPF::fromBufferSwitch(buffer,&clientType);
    nbytes += MPF::fromBufferSwitch(buffer,&clientExtra);
    return(nbytes);
}

void Message:: print() const
{
    printf("type                 %d\n",type);
    printf("clientType           %d\n",clientType);
    printf("clientExtra          %d\n",clientExtra);
    printf("type                 %d\n",type);
    printf("pMessageClient       %p\n",pvt->pMessageClient);
    printf("pMessageRouterClient %p\n",pvt->pMessageRouterClient);
    printf("pMessageRouterServer %p\n",pvt->pMessageRouterServer);
    printf("pMessageServer       %p\n",pvt->pMessageServer);
    printf("routerClientPvt      %p\n",pvt->routerClientPvt);
    printf("routerServerPvt      %p\n",pvt->routerServerPvt);
}

// MessageClient
struct MessageClientPrivate {
    MessageRouterClient	*pMessageRouterClient;
    MessageRouterServer *pMessageRouterServer;
    MessageServer	*pMessageServer;
    void		*routerServerPvt;
    WatchDog		*wdId;
};

MessageClient:: MessageClient(clientCallback callback,void *clientpvt)
: clientCall(callback), routerPvt(0), clientPvt(clientpvt)
{
    pvt = new MessageClientPrivate;
    ::memset((char *)pvt,0,sizeof(MessageClientPrivate));
    pvt->wdId = new WatchDog;
}

int MessageClient::bind(char *server, int location)
{
    MessageRouterClient *pMessageRouterClient;

    pMessageRouterClient = prouter->findMessageRouterClient(location);
    if(!pMessageRouterClient)  {
	ConnectMessage *pmessage = new ConnectMessage;
	MessagePvt	*msgPvt = pmessage->pvt;

	pvt->pMessageRouterClient = 0;
	msgPvt->pMessageClient = this;
	pmessage->status = connectNoRouter;
	bindCallback(pmessage);
	return(-1);
    }
    pvt->pMessageRouterClient = pMessageRouterClient;
    pMessageRouterClient->bind(server,this);
    return(0);
}

int MessageClient::send(Message *message)
{
    MessagePvt		*msgPvt = (MessagePvt *)message->pvt;

    msgPvt->pMessageClient = this;
    if(!pvt->pMessageRouterClient) return(-1);
    MessageRouterClient *pMessageRouterClient = pvt->pMessageRouterClient;
    msgPvt->pMessageRouterClient = pMessageRouterClient;
    msgPvt->pMessageRouterServer = pvt->pMessageRouterServer;
    msgPvt->pMessageServer = pvt->pMessageServer;
    msgPvt->routerServerPvt = pvt->routerServerPvt;
    return(pMessageRouterClient->send(message));
}

void MessageClient::bindCallback(ConnectMessage *message)
{
    MessagePvt	*msgPvt = message->pvt;

    switch(message->status) {
        case connectNo:
        case connectNoServer:
        case connectNoRouter:
	    pvt->pMessageRouterServer = 0;
	    pvt->pMessageServer = 0;
	    pvt->routerServerPvt = 0;
	    break;
	case connectYes:
	    pvt->pMessageRouterClient = msgPvt->pMessageRouterClient;
	    pvt->pMessageRouterServer = msgPvt->pMessageRouterServer;
	    pvt->pMessageServer = msgPvt->pMessageServer;
	    pvt->routerServerPvt = msgPvt->routerServerPvt;
	    break;
	case connectQueueFull:
	    printf("MessageClient::bindCallback connectQueueFull logic error\n");
	    break;
    }
    if(clientCall) (*clientCall)(message,clientPvt);
}

// MessageServer

struct MessageServerPrivate {
    DLList *pmessageList;
    SEM_ID block;
    char *name;
    int queueSize;
    // statistics
    unsigned long queueRequests;
    unsigned long queueFullResponses;
    unsigned long replyRequests;
};

MessageServer:: MessageServer(const char* name, int queueSize)
{
    pvt = new MessageServerPrivate;
    pvt->pmessageList = new DLList;
    pvt->block = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
    pvt->name = new char[::strlen(name)+1];
    ::strcpy(pvt->name,name);
    pvt->queueSize = (queueSize<=0) ? defaultServerQueueSize : queueSize;
    pvt->queueRequests = 0;
    pvt->queueFullResponses = 0;
    pvt->replyRequests = 0;
    prouter->registerServer(this);
}

void MessageServer:: waitForMessage()
{
    if(pvt->pmessageList->getCount()>0) return;
    semTake(pvt->block,WAIT_FOREVER);
}

Message *MessageServer:: receive()
{
    DLNode *node;
    Message *pMessage = 0;

    pvt->pmessageList->lock();
    node = pvt->pmessageList->getFirst();
    if(node) {
        pMessage = (Message *)node->obj;
        pvt->pmessageList->freeNode(node);
    }
    pvt->pmessageList->unlock();
    return(pMessage);
}

Message *MessageServer:: allocReplyMessage(
    Message *clientMessage,messageType typ)
{
    Message *pMessage;
    MessagePvt *msgPvt;
    MessagePvt *clientmsgPvt;

    pMessage = Message::allocate(typ);
    if(!pMessage) return(0);
    pMessage->clientType = clientMessage->clientType;
    pMessage->clientExtra = clientMessage->clientExtra;
    msgPvt = pMessage->pvt;
    clientmsgPvt = clientMessage->pvt;
    ::memcpy((char *)msgPvt,(char *)clientmsgPvt,sizeof(MessagePvt));
    return(pMessage);
}

void MessageServer::queueMessage(Message *pMessage)
{
    DLNode *node;

    ++pvt->queueRequests;
    if(pvt->queueSize>0 && (pvt->pmessageList->getCount() >=pvt->queueSize)) {
	++pvt->queueFullResponses;
	ConnectMessage *preply = 
	    (ConnectMessage *)allocReplyMessage(pMessage,messageTypeConnect);
	preply->status = connectQueueFull;
	reply(preply);
	delete pMessage;
	return;
    }
    pvt->pmessageList->lock();
    node = DLList::allocNode();
    node->obj = (void *)pMessage;
    pvt->pmessageList->addEnd(node);
    pvt->pmessageList->unlock();
    semGive(pvt->block);
}

int MessageServer::reply(Message *pMessage)
{
    ++pvt->replyRequests;
    MessagePvt *msgPvt = pMessage->pvt;
    MessageRouterServer *pMessageRouterServer = msgPvt->pMessageRouterServer;
    int	status=-1;
    if(pMessageRouterServer) status = pMessageRouterServer->reply(pMessage);
    return(status);
}

void MessageServer::setQueueSize(int size)
{
    pvt->queueSize = size;
}

const char *MessageServer::getName() const
{
    return((const char *)pvt->name);
}
    
void MessageServer::report() const
{
    printf("%s\n",pvt->name);
    printf("          queueSize %d\n", pvt->queueSize);
    printf("            inQueue %d\n",(pvt->pmessageList->getCount()));
    printf("      queueRequests %lu\n",pvt->queueRequests);
    printf(" queueFullResponses %lu\n",pvt->queueFullResponses);
    printf("      replyRequests %lu\n",pvt->replyRequests);
}

// Router.cc
struct RouterPvt {
    DLList	*clientRouterList;
    DLList	*serverRouterList;
    DLList	*serverList;
    SEM_ID	lock;
};

struct RouterClientRouterNode {
    int 		location;
    MessageRouterClient *pMessageRouterClient;
};

struct RouterServerRouterNode {
    int 		location;
    MessageRouterServer *pMessageRouterServer;
};

int mrr()
{
    prouter->routerReport();
    return(0);
}

int msr(const char *name)
{
    prouter->serverReport(name);
    return(0);
}

int routerInit()
{
    if(prouter) {
	printf("routerInit: Router already exists\n");
	return(-1);
    }
    prouter = new Router;
    return(0);
}

Router::Router()
{
    pvt = new RouterPvt;
    pvt->clientRouterList = new DLList;
    pvt->serverRouterList = new DLList;
    pvt->serverList = new DLList;
    pvt->lock = semBCreate(SEM_Q_FIFO, SEM_FULL);
}

void Router::routerReport()
{
    DLNode	*node;
    RouterClientRouterNode *pRouterClientRouterNode;
    RouterServerRouterNode *pRouterServerRouterNode;

    node = pvt->clientRouterList->getFirst();
    printf("clientRouterList\n");
    while(node) {
	pRouterClientRouterNode = (RouterClientRouterNode *)node->obj;
	MessageRouterClient *pMessageRouterClient =
	    pRouterClientRouterNode->pMessageRouterClient;
	pMessageRouterClient->report();
	node = pvt->clientRouterList->getNext(node);
    }
    node = pvt->serverRouterList->getFirst();
    printf("serverRouterList\n");
    while(node) {
	pRouterServerRouterNode = (RouterServerRouterNode *)node->obj;
	MessageRouterServer *pMessageRouterServer =
	    pRouterServerRouterNode->pMessageRouterServer;
	pMessageRouterServer->report();
	node = pvt->serverRouterList->getNext(node);
    }
}

void Router::serverReport(const char *name) 
{
    DLNode *node = pvt->serverList->getFirst();
    while(node) {
	MessageServer *pMessageServer = (MessageServer *)node->obj;
	if(!name || ::strcmp(pMessageServer->getName(),name)==0) 
	    pMessageServer->report();
	node = pvt->serverList->getNext(node);
    }
}

int Router::registerClientRouter(
    MessageRouterClient *pMessageRouterClient, int location)
{
    DLNode *node = DLList::allocNode();
    RouterClientRouterNode *pRouterClientRouterNode;

    pRouterClientRouterNode = new RouterClientRouterNode;
    pRouterClientRouterNode->pMessageRouterClient = pMessageRouterClient;
    pRouterClientRouterNode->location = location;
    node->obj = (void *)pRouterClientRouterNode;
    semTake(pvt->lock,WAIT_FOREVER);
    pvt->clientRouterList->addEnd(node);
    semGive(pvt->lock);
    return(0);
}


MessageRouterClient *Router::findMessageRouterClient(int location)
{
    DLNode *node;
    RouterClientRouterNode *pRouterClientRouterNode;

    semTake(pvt->lock,WAIT_FOREVER);
    node = pvt->clientRouterList->getFirst();
    while(node){
	pRouterClientRouterNode = (RouterClientRouterNode *)node->obj;
	if(pRouterClientRouterNode->location == location) {
	    semGive(pvt->lock);
	    return(pRouterClientRouterNode->pMessageRouterClient);
	}
	node = pvt->clientRouterList->getNext(node);
    }
    semGive(pvt->lock);
    return(0);
}

int Router::registerServerRouter(
    MessageRouterServer *pMessageRouterServer, int location)
{
    DLNode *node = DLList::allocNode();
    RouterServerRouterNode *pRouterServerRouterNode;

    pRouterServerRouterNode = new RouterServerRouterNode;
    pRouterServerRouterNode->pMessageRouterServer = pMessageRouterServer;
    pRouterServerRouterNode->location = location;
    node->obj = (void *)pRouterServerRouterNode;
    semTake(pvt->lock,WAIT_FOREVER);
    pvt->serverRouterList->addEnd(node);
    semGive(pvt->lock);
    return(0);
}

MessageRouterServer *Router::findMessageRouterServer(int location)
{
    DLNode *node;
    RouterServerRouterNode *pRouterServerRouterNode;

    semTake(pvt->lock,WAIT_FOREVER);
    node = pvt->serverRouterList->getFirst();
    while(node){
	pRouterServerRouterNode = (RouterServerRouterNode *)node->obj;
	if(pRouterServerRouterNode->location == location) {
	    semGive(pvt->lock);
	    return(pRouterServerRouterNode->pMessageRouterServer);
	}
	node = pvt->serverRouterList->getNext(node);
    }
    semGive(pvt->lock);
    return(0);
}

int Router::registerServer(MessageServer *pMessageServer)
{
    DLNode *node = DLList::allocNode();

    node->obj = (void *)pMessageServer;
    semTake(pvt->lock,WAIT_FOREVER);
    pvt->serverList->addEnd(node);
    semGive(pvt->lock);
    return(0);
}

MessageServer *Router::findMessageServer(const char *name)
{
    DLNode *node;
    MessageServer *pMessageServer;

    semTake(pvt->lock,WAIT_FOREVER);
    node = pvt->serverList->getFirst();
    while(node){
	pMessageServer = (MessageServer *)node->obj;
	if(::strcmp(pMessageServer->getName(),name)==0) {
	    semGive(pvt->lock);
	    return(pMessageServer);
	}
	node = pvt->serverList->getNext(node);
    }
    semGive(pvt->lock);
    return(0);
}

void *MPF::allocate(DLList **ppFreeList,size_t size)
{
    Message *pmessage = 0;
    DLList  *pFreeList;
    DLNode  *node;

    if(*ppFreeList == 0) *ppFreeList = new DLList;
    pFreeList = *ppFreeList;
    pFreeList->lock();
    node = pFreeList->getFirst();
    if(node) {
	pmessage = (Message *)node->obj;
	pFreeList->freeNode(node);
    }
    pFreeList->unlock();
    if(!pmessage){
	pmessage = (Message *)calloc(1,size);
	pmessage->pvt = (MessagePvt *)calloc(1,sizeof(MessagePvt));
    }
    ::memset((char *)pmessage->pvt,0,sizeof(MessagePvt));
    return(pmessage);
}

// size is not needed. Included to match delete
void MPF::free(DLList *pFreeList,void *p,size_t )
{
    DLNode  *node;
    Message *pmessage = (Message *)p;

    node = DLList::allocNode();
    node->obj = pmessage;
    pFreeList->lock();
    pFreeList->addEnd(node);
    pFreeList->unlock();
}

int32 MPF::toBuffer(char **buffer,const char *pstring)
{
    int32 nbytes = strlen(pstring) + 1;

    toBuffer(buffer,nbytes);
    memcpy(*buffer,pstring,nbytes);
    *buffer += nbytes;
    return(nbytes +4);
}

int32 MPF::fromBuffer(const char **buffer,char *pstring,int32 maxlen)
{
    int32 nbytes=0;

    fromBuffer(buffer,&nbytes);
    if(nbytes>=maxlen) {
	memcpy(pstring,*buffer,maxlen-1);
	pstring[maxlen]=0;
    } else {
        memcpy(pstring,*buffer,nbytes);
    }
    *buffer += nbytes;
    return(nbytes +4);
}

int32 MPF::fromBufferSwitch(const char **buffer,char *pstring,int32 maxlen)
{
    int32 nbytes=0;

    fromBufferSwitch(buffer,&nbytes);
    if(nbytes>=maxlen) {
	memcpy(pstring,*buffer,maxlen-1);
	pstring[maxlen]=0;
    } else {
        memcpy(pstring,*buffer,nbytes);
    }
    *buffer += nbytes;
    return(nbytes +4);
}

int32 MPF::toBuffer(char **buffer,const void *data,int32 nbytes)
{
    memcpy(*buffer,(char *)data,nbytes);
    *buffer += nbytes;
    return(nbytes);
}

int32 MPF::fromBuffer(const char **buffer,void *data,int32 nbytes)
{
    memcpy((char *)data,*buffer,nbytes);
    *buffer += nbytes;
    return(nbytes);
}
