// WatchDog.cc

/* Author:  Marty Kraimer Date: 26JUN1998 */
 
/********************COPYRIGHT NOTIFICATION**********************************
This software was developed under a United States Government license
described on the COPYRIGHT_UniversityOfChicago file included as part
of this distribution.
****************************************************************************/

#include <vxWorks.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <wdLib.h>
#include <rngLib.h>
#include <semLib.h>
#include <intLib.h>
#include <taskLib.h>

#include "WatchDog.h"

class WatchDogPrivate
{
public:
    WatchDogPrivate();
    ~WatchDogPrivate();
    void start(int delaySeconds,watchDogCallback callback,void *userPvt);
    void cancel();
    WDOG_ID wdId;
    watchDogCallback userCallback;
    void *userPvt;
};

int watchDogQueueSize = 1000;
static bool initialized=false;
static SEM_ID watchDogSem;
static RING_ID watchDogQ;
static int watchDogTaskId;

static void watchDogPrivateCallback(WatchDogPrivate *pwatchDogPrivate);
static void watchDogTask(void);
int watchDogSetQueueSize(int size);
static void initWatchDog(void);

WatchDog::WatchDog()
{
    WatchDogPrivate *pwatchDogPrivate = new WatchDogPrivate;
    wdPvt = (void *)pwatchDogPrivate;
}
WatchDog:: ~WatchDog()
{
    WatchDogPrivate *pwatchDogPrivate = (WatchDogPrivate *)wdPvt;
    delete pwatchDogPrivate;
}
void WatchDog::start(int delay,watchDogCallback callback,void *userPvt)
{
    WatchDogPrivate *pwatchDogPrivate = (WatchDogPrivate *)wdPvt;
    pwatchDogPrivate->start(delay,callback,userPvt);
}
void WatchDog::cancel()
{
    WatchDogPrivate *pwatchDogPrivate = (WatchDogPrivate *)wdPvt;
    pwatchDogPrivate->cancel();
}

WatchDogPrivate::WatchDogPrivate()
{
    if(!initialized) {
	initWatchDog();
	initialized = true;
    }
    wdId = wdCreate();
    if(wdId==0) {
	printf("WatchDogPrivate create failed\n");
	taskSuspend(0);
    }
}

WatchDogPrivate:: ~WatchDogPrivate()
{
    STATUS status = wdDelete(wdId);	
    if(status==ERROR) {
	printf("WatchDogPrivate create failed\n");
	taskSuspend(0);
    }
}

void WatchDogPrivate::start(int delay,watchDogCallback callback,void *userPvt)
{
    userCallback = callback;
    this->userPvt = userPvt;
    STATUS status = wdStart(wdId,delay*sysClkRateGet(),
	(FUNCPTR)watchDogPrivateCallback,(int)this);
    if(status==ERROR) {
	logMsg("WatchDogPrivate start failed\n");
	taskSuspend(0);
    }
}

void WatchDogPrivate::cancel()
{
    STATUS status = wdCancel(wdId);
    if(status==ERROR) {
	logMsg("WatchDogPrivate cancel failed\n");
	taskSuspend(0);
    }
}

// watchDogPrivateCallback called at interrupt level
static void watchDogPrivateCallback(WatchDogPrivate *pwatchDogPrivate)
{
    static int newOverflow=TRUE;
    int lockKey;
    int n;

    lockKey = intLock();
    n = rngBufPut(watchDogQ,(char *)&pwatchDogPrivate,sizeof(pwatchDogPrivate));
    intUnlock(lockKey);
    if(n!=sizeof(pwatchDogPrivate)) {
	if(newOverflow)logMsg("rngBufPut overflow in wdCallback\n");
	newOverflow = FALSE;
    } else {
	newOverflow = TRUE;
    }
    semGive(watchDogSem);
}

static void watchDogTask(void)
{
    WatchDogPrivate *pwatchDogPrivate=0;
    while(true) {
	if(semTake(watchDogSem,WAIT_FOREVER)!=OK)
	    printf("watchDogTask semTake returned error\n");
	while (rngNBytes(watchDogQ)>=(int)sizeof(pwatchDogPrivate)){
	    if(rngBufGet(watchDogQ,(char *)&pwatchDogPrivate,
	    sizeof(pwatchDogPrivate))!=(int)sizeof(pwatchDogPrivate)) {
		printf("watchDogTask rngBufGet returned error\n");
	    } else {
	        (*pwatchDogPrivate->userCallback)(pwatchDogPrivate->userPvt);

	    }
	}
    }
}

int watchDogSetQueueSize(int size)
{
    watchDogQueueSize = size;
    return(0);
}

static char taskname[] = "watchDog";

static void initWatchDog(void)
{
    if((watchDogQ = rngCreate(sizeof(void *) * watchDogQueueSize))==NULL){
        printf("WatchDog: initWatchDog failed\n");
        exit(1);
    }
    if((watchDogSem=semBCreate(SEM_Q_FIFO,SEM_EMPTY))==NULL)
        printf("semBcreate failed in initWatchDog\n");
    watchDogTaskId = taskSpawn(taskname,130,VX_FP_TASK,2000,
        (FUNCPTR)watchDogTask,
        0,0,0,0,0,0,0,0,0,0);
}
